library(googledrive)
library(dplyr, warn.conflicts = FALSE)
library(openxlsx)
library(RPostgreSQL)

# drive_auth(reset = TRUE)
# setwd("~/git/non_gaap")

# Download files ----
flist <- 
    drive_ls(path = "Quarterly non-GAAP Updates") %>% 
    select(name) %>% 
    arrange(desc(name)) %>% 
    unlist()

lapply(flist, drive_download)

# Upload files ----
upload_data <- function(fname, table) {
    # This function uploads xlsx data to pg
    data <- as_tibble(read.xlsx(fname, sheet = 1, check.names = T)) 
    
    # Unify total column number, the valid number is 27. Extra columns are meaningless.
    if (ncol(data) < 27) {
        data <- 
            data %>% 
            mutate(non_current_non_gaap = NA_character_)
    } else if (ncol(data) > 27) {
        data <- 
            data %>% 
            select(1:27)
    }
    
    print(paste0("Start to import ", fname))
    print(dim(data))
    
    # Unify column names. Each file has slightly different column names
    names(data) <- c("company", "cik_code", "ticker", "market", "sic_code", "sic_description", "filing", "date", "fiscal_period", 
                     "name_of_the_metric", "metric_key", "metric_type", "line", "aa_taxonomy", "line_sequence_number", "dollar_value", 
                     "number_of_metrics", "total_lines", "type_of_line", "not_clearly_marked_as_non_gaap_not_reconciled", 
                     "reconciled_from_multiple_items", "new_reconciliation_line", "new_metric", "company_specific_item", 
                     "new_non_gaap_methodology", "notes", "non_current_non_gaap" )
    data <- 
        data %>% 
        mutate_at(.vars = vars(cik_code, sic_code, number_of_metrics, total_lines, not_clearly_marked_as_non_gaap_not_reconciled, 
                               reconciled_from_multiple_items, new_reconciliation_line, new_metric), 
                  .funs = funs(as.integer(.))) %>% 
        mutate(date = as.Date(date %>% as.integer(), origin="1899-12-30"), 
               new_non_gaap_methodology = as.character(new_non_gaap_methodology), 
               company_specific_item = as.character(company_specific_item), 
               non_current_non_gaap = as.character(non_current_non_gaap)) %>% 
        # Fix error
        filter(!(dollar_value %in% c("No longer filing", "No flling", "no filing", "No 2.02", "No filing", "No 8-K"))) %>% 
        mutate(dollar_value = if_else(dollar_value == "-", NA_real_, dollar_value %>% as.double())) %>% 
        filter(!is.na(cik_code), !is.na(company))
    
    dbWriteTable(pg, table, data, append = TRUE, row.names = FALSE, overwrite = FALSE)
    # dbDisconnect(pg)
    print("Import ended.")
}

pg <- dbConnect(PostgreSQL())
proj_schema <- "non_gaap"
rs <- dbGetQuery(pg, paste0("SET search_path = '", proj_schema, "'"))

dbGetQuery(pg, "DROP TABLE IF EXISTS non_gaap.detail")
sapply(flist, upload_data, "detail")
# upload_data(fname, table)

dbGetQuery(pg, "ALTER TABLE non_gaap.detail OWNER TO non_gaap")
dbGetQuery(pg, "GRANT SELECT ON non_gaap.detail TO non_gaap_access")
db_comment <- paste0("CREATED USING create_detail.R IN GitHub azakolyukina/non_gaap ON ", Sys.time())
dbGetQuery(pg, sprintf("COMMENT ON TABLE non_gaap.detail IS '%s';", db_comment))
dbDisconnect(pg)
