# This code creates data for estimation

# Remove everything
#rm(list = ls())

library(RPostgreSQL)
library(dplyr)
library(tidyr)
library(psych)

pg <- dbConnect(PostgreSQL())

# Define PG data ----

# Project schema
rs <- dbGetQuery(pg, "SET search_path TO non_gaap, public")
firms <- tbl(pg, "firms")
ibes_mb <- tbl(pg, "ibes_mb")
executives <- tbl(pg, "executives")
ccmxpf_linktable <- tbl(pg, sql("SELECT * FROM crsp.ccmxpf_linktable"))

# Set params ----

late_date <- "2100-01-01"

# Pre-process data ----

# Get gvkey for ibes_mb data
ibes_mb_alt <- 
    ibes_mb %>%
    left_join(ccmxpf_linktable %>%
              mutate(linkenddt = coalesce(linkenddt, late_date)) %>%
              filter(linkprim %in% c("P", "C"), usedflag == 1L) %>%
              rename(permno = lpermno)) %>%
    # Keep valid links
    filter(linkdt < pends & pends <= linkenddt) %>%
    filter(!is.na(gvkey)) %>%
    mutate(year = sql("date_part('year', pends)"), 
           month = sql("date_part('month', pends)")) %>%
    select(gvkey, year, month, consens_n, ce_avg) %>%
    distinct() %>%
    compute()

executives_alt <- 
    executives %>%
    select(company_id, executive_id, gvkey, 
           fyearq = fyear, fqtr, mmratio, delta_vested) %>%
    distinct() %>%
    compute()

# Write to PG ----

dbGetQuery(pg, "DROP TABLE IF EXISTS est_merged")

# Require non-missing variables used in moments computatation
est_merged <- 
    firms %>%
    mutate(year = sql("date_part('year', datadate)"), 
           month = sql("date_part('month', datadate)")) %>%
    left_join(ibes_mb_alt) %>%
    left_join(executives_alt) %>%
    select(-year, -month) %>%
    filter(!is.na(gvkey) & !is.na(fyearq) & !is.na(fqtr)) %>%
    copy_to(pg, ., "est_merged", temporary = FALSE, 
            indexes = list(c("gvkey", "fyearq", "fqtr"), 
                           c("gvkey", "fqtr"), 
                           c("gvkey")))

comment <- 'CREATED USING azakolyukina/non_gaap'
db_comment <- paste0("COMMENT ON TABLE est_merged IS '",
                     comment, " ON ", Sys.time() , "'; ") 
dbGetQuery(pg, db_comment)

dbGetQuery(pg, "ALTER TABLE est_merged OWNER TO non_gaap")
dbGetQuery(pg, "GRANT SELECT ON TABLE est_merged TO non_gaap_access")
dbGetQuery(pg, paste0("VACUUM ANALYZE est_merged"))

dbDisconnect(pg)
