# This code creates data for estimation

# Remove everything
#rm(list = ls())

library(RPostgreSQL)
library(dplyr)
library(tidyr)
library(psych)

pg <- dbConnect(PostgreSQL())

# Define PG data ----

# Project schema
rs <- dbGetQuery(pg, "SET search_path TO non_gaap, public")
est_merged <- tbl(pg, "est_merged")

# Define params ----

q_bea_at_cutoff <- 0.10 # cutoff for intang. capital intensity
# ce_avg = (actual_adj - consens_avg)/prc_adj20
mb_cutoff <- 0.0005 # cutoff on just meet-or-beat  

# Pre-process variables and write table ----

dbGetQuery(pg, "DROP TABLE IF EXISTS est_set")

# Require non-missing variables used in moments computatation
est_set <- 
    est_merged %>%
    # Keep bcgw firms only
    filter(bcgw_start <= datadate & datadate <= bcgw_end) %>%
    select(-bcgw_start, -bcgw_end) %>%
    # Exclude regulated utilities and non-operating establ. as in Peters and Taylor (2017)
    # WARNING: Eisfeldt and Papanikolaou (2014) keep finance. Peters and Taylor (2017) drop finance
    filter(!is.na(ind_ep6) & ind_ep6 != "Utilities" & ind_ep6 != "Finance" & ind_ep6 != "Non-operating est.") %>%
    # Exclude missing at, missing sales, at < 5 mil as in Peters and Taylor (2017)
    filter(!is.na(at) & !is.na(sale) & at > 5) %>%
    # Keep only above intangible intensity cutoff by BEA criterion
    group_by(gvkey) %>%
    mutate(q_bea_at_min = min(q_bea_at)) %>%
    ungroup() %>%
    filter(q_bea_at_min >= q_bea_at_cutoff) %>%
    select(-q_bea_at_min) %>%
    # Keep from the FIRST datadate of non-GAAP adj. reporting
    mutate(datadate_ng = if_else(mgr_exclude == 1, datadate, '2100-01-01')) %>%
    group_by(gvkey) %>%
    mutate(is_ever_ng = max(mgr_exclude)) %>%
    filter(datadate >= min(datadate_ng)) %>%
    ungroup() %>%
    select(-datadate_ng) %>%
    # Require non-missing variables
    # WARNING: check colSums(is.na(DATA)) before filtering
    filter_at(.vars = vars(-mgr_exclude, # non-GAAP indicator
                           -consens_n, -ce_avg, -matches("ibes"), # IBES data
                           -company_id, -executive_id, -mmratio, -delta_vested, # Equilar data
                           -matches("ppent")), # Physical capital intensity
              all_vars(!is.na(.))) %>%
    collect() %>%
    # Winsorize at 1-/99-
    mutate_at(.vars = vars(-gvkey, -datadate, -fyearq, -fqtr, -sich, -ind_ep6,
                           -is_ever_ng, -consens_n, -company_id, -executive_id,
                           # Do NOT winsorize growth since restricted to [-2; 2]
                           # by construction
                           -matches("_gr1")), 
              .funs = funs(winsor(., 0.01))) %>%
    # Require at least T = 2 per gvkey-fqtr, i.e., 5 years of data with rho_X
    group_by(gvkey, fqtr) %>%
    mutate(n_obs = n()) %>%
    ungroup() %>%
    filter(n_obs >= 2) %>%
    # Create indicators for splits
    # Meet-or-beat: {mb0, mb1, mb2}
    mutate(mb0 = if_else(consens_n > 1 & ce_avg < 0, 1, 0), 
           mb1 = if_else(consens_n > 1 & ce_avg >= 0 & ce_avg < mb_cutoff, 1, 0), 
           mb2 = if_else(consens_n > 1 & ce_avg >= mb_cutoff, 1, 0)) %>%
    # Edmans et al. vested measure mmratio split
    mutate(mmratio0 = (mmratio == 0), 
           mmratio1 = (mmratio > 0)) %>%
    # Size, market-to-book, and delta_vested terciles
    group_by(gvkey) %>%
    mutate_at(.vars = vars(mcap, market_to_book, delta_vested, q_bea_ppent, consens_n), 
              .funs = funs(avg = mean(., na.rm = TRUE))) %>%
    ungroup() %>%
    mutate_at(.vars = vars(mcap_avg, market_to_book_avg, delta_vested_avg, q_bea_ppent_avg, consens_n_avg),
              .funs = funs(p30 = quantile(., 0.30, na.rm = TRUE),
                           p70 = quantile(., 0.70, na.rm = TRUE))) %>%
    mutate(mcap_low = if_else(mcap_avg <= mcap_avg_p30, 1, 0),
           mcap_high = if_else(mcap_avg >= mcap_avg_p70, 1, 0),
           market_to_book_low = if_else(market_to_book_avg <= market_to_book_avg_p30, 1, 0),
           market_to_book_high = if_else(market_to_book_avg >= market_to_book_avg_p70, 1, 0),
           delta_vested_low = if_else(delta_vested_avg <= delta_vested_avg_p30, 1, 0),
           delta_vested_high = if_else(delta_vested_avg >= delta_vested_avg_p70, 1, 0),
           q_bea_ppent_low = if_else(q_bea_ppent_avg <= q_bea_ppent_avg_p30, 1, 0),
           q_bea_ppent_high = if_else(q_bea_ppent_avg >= q_bea_ppent_avg_p70, 1, 0),
           consens_n_low = if_else(consens_n_avg <= consens_n_avg_p30, 1, 0),
           consens_n_high = if_else(consens_n_avg >= consens_n_avg_p70, 1, 0)) %>%
    # WARNING: Prevent failure when writing a table
    select(-matches("_p30|_p70")) %>%
    arrange(gvkey, datadate) %>%
    copy_to(pg, ., "est_set", temporary = FALSE, 
            indexes = list(c("gvkey", "fyearq", "fqtr"), 
                           c("gvkey", "fqtr"), 
                           c("gvkey")))

comment <- 'CREATED USING azakolyukina/non_gaap/create_est_set.R'
db_comment <- paste0("COMMENT ON TABLE est_set IS '",
                     comment, " ON ", Sys.time() , "'; ") 
dbGetQuery(pg, db_comment)

dbGetQuery(pg, "ALTER TABLE est_set OWNER TO non_gaap")
dbGetQuery(pg, "GRANT SELECT ON TABLE est_set TO non_gaap_access")
dbGetQuery(pg, "VACUUM ANALYZE est_set")

dbDisconnect(pg)
