# This code computes fundamental variables from Compustat

#.rs.restartR()
#(.packages())
# rm(list = ls())

library(RPostgreSQL)
library(dplyr)

pg <- dbConnect(PostgreSQL())

# Define PG data ----

# Project schema
rs <- dbGetQuery(pg, "SET search_path TO non_gaap")
intang <- tbl(pg, "intang")
prelim <- tbl(pg, "prelim")
bcgw <- tbl(pg, "bcgw")
value <- tbl(pg, "value")
cpi <- tbl(pg, "cpi")
ibes_mb <- tbl(pg, "ibes_mb")

# Outside of project schema
fundq <- tbl(pg, sql("SELECT * FROM comp.fundq"))
ccmxpf_linktable <- tbl(pg, sql("SELECT * FROM crsp.ccmxpf_linktable"))

# Set params ----

start_date <- "1998-01-01"
end_date <- "2100-01-01"

# Collect non-GAAP EPS ----

bcgw_alt <- 
    bcgw %>%
    mutate(caldt_year = sql("date_part('year', datadate)"), 
           caldt_month = sql("date_part('month', datadate)")) %>%
    # min/max date for gvkey
    group_by(gvkey) %>%
    mutate(bcgw_start = min(datadate), 
           bcgw_end = max(datadate)) %>%
    ungroup() %>%
    select(gvkey, datadate, bcgw_start, bcgw_end, caldt_year, caldt_month, mgr_exclude, ng_eps = mgr_ng_eps) %>%
    compute()

# Collect value and adjust by end-of-the-month CPI ----

value_alt <- 
    value %>%
    # CRSP date for market capitalization and date_lag for lagged market capitalization
    mutate(caldt_year = sql("date_part('year', date)"), 
           caldt_month = sql("date_part('month', date)"), 
           caldt_year_lag = sql("date_part('year', date_lag)"), 
           caldt_month_lag = sql("date_part('month', date_lag)")) %>%
    # Merge CPI
    left_join(cpi %>%
                  select(caldt_year, caldt_month, cpifac)) %>%
    # Merge CPI
    left_join(cpi %>%
                  select(caldt_year_lag = caldt_year, 
                         caldt_month_lag = caldt_month, 
                         cpifac_lag = cpifac)) %>%
    # Scale by CPI
    mutate(mcap_cpi = mcap*cpifac, 
           mcap_lag_cpi = mcap_lag*cpifac) %>%
    select(gvkey, datadate, mcap_cpi, mcap_lag_cpi) %>%
    compute()

# Collect quarterly fundamentals ----

# Acct. values are in millions
colsq <- c("gvkey", "datadate", "fyearq", "fqtr",
          "atq", # Assets - Total 
          "saleq", # Sales/Turnover (Net)
          "ppentq", # PPENTQ -- Property Plant and Equipment - Total (Net)
          "epsfiq", # EPSFIQ -- Earnings Per Share (Diluted) - Including Extraordinary Items
          "cshfdq", # CSHFDQ -- Com Shares for Diluted EPS
          "cshprq", # CSHPRQ -- Common Shares Used to Calculate Earnings Per Share - Basic
          "cshoq", # Common Shares Outstanding (in millions of shares)
          "oancfy", # Operating Activities - Net Cash Flow
          "capxy", # CAPXY -- Capital Expenditures
          "sppey", # SPPEY -- Sale of Property
          "seqq", # SEQ -- Stockholders' Equity - Total
          "txditcq", #TXDITC -- Deferred Taxes and Investment Tax Credit
          "pstkq", # PSTK -- Preferred/Preference Stock (Capital) - Total
          "prccq", # PRCCQ -- Price Close - Quarter
          "adjex" # Cumulative Adjustment Factor by Ex-Date - Fiscal (adjust PRC by dividing by CFACPR )
)

# Quarterly fundamentals
fundq_alt <- 
    fundq %>%
    filter(datadate >= start_date & datadate <= end_date) %>%
    filter(indfmt == 'INDL' & datafmt == 'STD' & popsrc == 'D' & consol == 'C' & fic == 'USA') %>%
    select(one_of(colsq)) %>%
    mutate(caldt_year = sql("date_part('year', datadate)"), 
           caldt_month = sql("date_part('month', datadate)")) %>%
    filter(!is.na(datadate) & !is.na(fyearq) & !is.na(fqtr)) %>%
    # For duplicate {gvkey, datadate}, keep min(fyearq)
    group_by(gvkey, datadate) %>%
    filter(fyearq == min(fyearq)) %>%
    ungroup() %>%
    # For remaining duplicate {gvkey, datadate}, keep min(fqtr)
    group_by(gvkey, datadate) %>%
    filter(fqtr == min(fqtr)) %>%
    ungroup() %>%
    # Convert year-to-date to quarterly for {capxy, sppey, oancfy}
    group_by(gvkey, fyearq) %>%
    arrange(datadate) %>%
    mutate(oancfq = if_else(fqtr != 1, oancfy - lag(oancfy), oancfy), 
           capxq = if_else(fqtr != 1, capxy - lag(capxy), capxy), 
           sppeq = if_else(fqtr != 1, sppey - lag(sppey), sppey)) %>%
    ungroup() %>%
    # Merge with preliminary history
    left_join(prelim %>%
                  select(gvkey, datadate, epsfiq_prelim = epsfiq_p)) %>%
    mutate(epsfiq = coalesce(epsfiq_prelim, epsfiq)) %>%
    compute()

# WARNING: do NOT use year-to-date amount since they close a trivial
# number of missing obs.
fundq_cpi <- 
    fundq_alt %>%
    # Merge non-GAAP EPS
    left_join(bcgw_alt %>%
                  select(-datadate)) %>%
    # Extrapolate bcgw_start/end dates
    group_by(gvkey) %>%
    mutate(bcgw_start = min(bcgw_start), 
           bcgw_end = max(bcgw_end)) %>%
    ungroup() %>%
    # Merge CPI
    left_join(cpi %>%
                  select(caldt_year, caldt_month, cpifac)) %>%
    # Compute real expenses
    mutate_at(.vars = vars(atq, saleq, ppentq, epsfiq, oancfq, capxq, sppeq, seqq, 
                           txditcq, pstkq, prccq, ng_eps), 
              .funs = funs(cpi = .*cpifac)) %>%
    # Merge firm value a la' Bradshaw and Sloan (2002)
    left_join(value_alt) %>%
    select(gvkey, datadate, fyearq, fqtr, matches("_cpi"), 
           cshfdq, cshprq, cshoq, adjex, cpifac, 
           bcgw_start, bcgw_end, mgr_exclude) %>%
    compute()

# Analyst-defined non-GAAP (Street earnings) -----

# Get gvkey for ibes_mb data
ibes_mb_alt <- 
    ibes_mb %>%
    left_join(ccmxpf_linktable %>%
                  mutate(linkenddt = coalesce(linkenddt, end_date)) %>%
                  filter(linkprim %in% c("P", "C"), usedflag == 1L) %>%
                  rename(permno = lpermno)) %>%
    # Keep valid links
    filter(linkdt < pends & pends <= linkenddt) %>%
    filter(!is.na(gvkey)) %>%
    mutate(year = sql("date_part('year', pends)"), 
           month = sql("date_part('month', pends)")) %>%
    select(gvkey, year, month, actual_adj) %>%
    distinct() %>%
    # Merge CPI
    left_join(cpi %>%
                  select(year = caldt_year, month = caldt_month, cpifac)) %>%
    mutate(actual_adj_cpi = actual_adj*cpifac) %>%
    compute()

# Compute variabes (scaled and growth) ----

fundq_calc <- 
    fundq_cpi %>%
    mutate(year = sql("date_part('year', datadate)"), 
           month = sql("date_part('month', datadate)")) %>%
    left_join(ibes_mb_alt) %>%
    # Merge intangible investment
    left_join(intang %>%
                  select(gvkey, datadate, fyearq, fqtr, sich, ind_ep6, matches("q_|delta_|w_"))) %>%
    # Set non-positive {atq, saleq} values to NA
    # WARNING: do NOT filter not to introduce gaps in time-series
    mutate_at(.vars = vars(atq_cpi, saleq_cpi), 
              .funs = funs(if_else(. <= 0, NA, .))) %>%
    # Replace missing values with 0s 
    mutate_at(.vars = vars(capxq_cpi, sppeq_cpi, txditcq_cpi, pstkq_cpi),
              .funs = funs(coalesce(., 0))) %>%
    # Fill in missing number of shares
    mutate(cshfdq = coalesce(cshfdq, cshprq, cshoq)) %>%
    # Compute vars for descriptives
    mutate(prc = if_else(adjex != 0, prccq_cpi/adjex, NA),
           book_equity = seqq_cpi + txditcq_cpi - pstkq_cpi) %>%
    mutate(market_to_book = (prc*cshoq + atq_cpi - book_equity)/atq_cpi,
           mcap = prc*cshoq, 
           at = atq_cpi, 
           sale = saleq_cpi, 
           vf = mcap_cpi, 
           vf_lag = mcap_lag_cpi) %>%
    # Compute intang. capital intensity
    # Intang. capital intensity relative to physical capital
    mutate_at(.vars = vars(matches("^q_")), 
              .funs = funs(at = ./atq_cpi, 
                           ppent = if_else(. != 0, ./NULLIF(. + ppentq_cpi, 0), 0))) %>%
    # Compute capital weighted depreciation rate (delta)
    mutate(delta_bea = (delta_org_bea*q_org_bea + delta_know_bea*q_know_bea)/NULLIF(q_bea, 0),
           delta_epw = (delta_org_epw*q_org_epw + delta_know_epw*q_know_epw)/NULLIF(q_epw, 0),
           delta_ep = (delta_org_ep*q_org_ep + delta_know_ep*q_know_ep)/NULLIF(q_ep, 0)) %>%
    # Compute a la' Bradshaw and Sloan (2002) value changes
    mutate(dvf = vf - vf_lag) %>%
    # Convert size vars to billions
    mutate_at(.vars = vars(mcap, at, sale), 
              .funs = funs(bn = ./(10**3))) %>%
    # Compute vars for the model 
    mutate(d = oancfq_cpi - (capxq_cpi - sppeq_cpi), # Free cash flows (d)
           pii = epsfiq_cpi*cshfdq, # GAAP earnings (pi)
           psi = ng_eps_cpi*cshfdq - epsfiq_cpi*cshfdq, # non-GAAP adjustment (psi)
           psi_ibes = actual_adj_cpi*cshfdq - epsfiq_cpi*cshfdq) %>% # # non-GAAP adjustment based on IBES (psi) 
    # Fill-in non-GAAP adj. in bcgw period with mgr_exclude = {0, 1}
    mutate(psi = if_else(!is.na(mgr_exclude), coalesce(psi, 0), psi), # Conservative (lost for X-rho)
           psi_ibes = if_else(!is.na(actual_adj_cpi), coalesce(psi_ibes, 0), psi_ibes)) %>% 
    # Compute non-GAAP earnings (piing)
    mutate(piing = pii + psi, 
           piing_ibes = pii + psi_ibes) %>%
    # Scale investment by capital
    mutate(w_q_bea = if_else(w_bea != 0, w_bea/NULLIF(q_bea, 0), 0), 
           w_q_epw = if_else(w_epw != 0, w_epw/NULLIF(q_epw, 0), 0), 
           w_q_ep = if_else(w_ep != 0, w_ep/NULLIF(q_ep, 0), 0), 
           w_q_noinit = if_else(w_noinit != 0, w_noinit/NULLIF(q_noinit, 0), 0)) %>%
    # Scale {d, pi, psi, piing} by capital
    mutate_at(.vars = vars(d, pii, psi, psi_ibes, piing, piing_ibes), 
              .funs = funs(q_bea = if_else(. != 0, ./NULLIF(q_bea, 0), 0), 
                           q_epw = if_else(. != 0, ./NULLIF(q_epw, 0), 0), 
                           q_ep = if_else(. != 0, ./NULLIF(q_ep, 0), 0), 
                           q_noinit = if_else(. != 0, ./NULLIF(q_noinit, 0), 0))) %>%
    group_by(gvkey, fqtr) %>%
    arrange(datadate) %>%
    # Compute iteract. terms for pricing moments
    mutate(pii2 = pii^2,
           psi2 = psi^2,
           psi_ibes2 = psi_ibes^2,
           piipsi = pii*psi,
           piipsi_ibes = pii*psi_ibes,
           piing2 = piing^2,
           piing_ibes2 = piing_ibes^2,
           piipiing = pii*piing,
           piipiing_ibes = pii*piing_ibes) %>%
    # Compute growth rates for {w, d, pii, psi, piing} for firm-quarters
    # Compute growth rates as in Stephen Terry's JMP
    # WARNING: cannot send . and ._lag to sql
    # Growth based on raw variables
    mutate_at(.vars = vars(w_bea, w_epw, w_ep, w_noinit, d, pii, psi, psi_ibes, dvf, 
                           pii2, psi2, psi_ibes2, piipsi, piipsi_ibes, piing, piing_ibes, piing2,  piing_ibes2, piipiing, piipiing_ibes),
              .funs = funs(gr1 = if_else(. == 0 & sql(lag(.)) == 0, 0, 
                                         2*(. - sql(lag(.)))/
                                             (sql(NULLIF(abs(1.0*.) + lag(abs(1.0*.)), 0)))))) %>%
    # Compute changes in {pii_q, psi_q, d_q, piing_q} for X-differences for rho_X for firm-quarters
    mutate_at(.vars = vars(matches("^pii_q|^psi_q|^psi_ibes_q|^d_q|^piing_q|^piing_ibes_q")), 
              .funs = funs(lag03 = . - lag(., 3), 
                           lag12 = lag(., 1) - lag(., 2))) %>%
    ungroup() %>%
    # Compute variables for pshuffle
    group_by(gvkey, fqtr) %>%
    arrange(datadate) %>%
    mutate(at_avg = coalesce((atq_cpi + lag(atq_cpi))/2, atq_cpi, lag(atq_cpi))) %>%
    mutate_at(.vars = vars(d, pii), 
              .funs = funs(at_avg  = ./at_avg)) %>%
    # Compute lag and lead for Dechow and Dichev measure
    mutate(d_at_avg_lag1 = lag(d_at_avg),
           d_at_avg_lead1 = lead(d_at_avg)) %>%
    ungroup() %>%
    compute()

# Write to PG ----

dbGetQuery(pg, "DROP TABLE IF EXISTS firms")

firms <- 
    fundq_calc %>%
    # Keep vars for estimation 
    select(gvkey, datadate, fyearq, fqtr, sich, ind_ep6, 
           bcgw_start, bcgw_end, mgr_exclude,
           market_to_book, mcap, mcap_bn, sale, sale_bn, at, at_bn, 
           saleq_cpi, ppentq_cpi, # Cobb-Douglas func. 
           d, pii, psi, vf, vf_lag, dvf, dvf_gr1, 
           pii2, pii2_gr1, psi2, psi2_gr1, piipsi, piipsi_gr1, 
           piing, piing_gr1, piing2, piing2_gr1, piipiing, piipiing_gr1,
           delta_bea, delta_epw, delta_ep,
           matches("^q_|^w_|^d_|^pii_|^psi_|^piing_|^piipsi_")) %>%
    compute(name = "firms", temporary = FALSE,
            indexes=list(c("gvkey", "datadate"), c("gvkey")))

comment <- 'CREATED USING azakolyukina/non_gaap/create_firms.R'
db_comment <- paste0("COMMENT ON TABLE firms IS '",
                     comment, " ON ", Sys.time() , "'; ") 
dbGetQuery(pg, db_comment)

dbGetQuery(pg, "ALTER TABLE firms OWNER TO non_gaap")
dbGetQuery(pg, "GRANT SELECT ON TABLE firms TO non_gaap_access")
dbGetQuery(pg, "VACUUM ANALYZE firms")

dbDisconnect(pg)
