# This code creates intangible capital and investment at quarterly frequency. Assumptions
# for the perpetual inventory method are from: 
#   (1) BEA-HH (_bea)
#   (2) Ewens, Peters, and Wang, WP (2019) "Acquisition Prices and the Measurement of Intangible Capital" (_ewp)
#   (3) Eisfeldt and Papanikolaou JF (2013) "Organization capital and the cross-section of expected returns" (_ep)

#.rs.restartR()
#(.packages())
# rm(list = ls())

library(RPostgreSQL)
library(dplyr)
library(tidyverse)
library(psych)

pg <- dbConnect(PostgreSQL())

# Define PG data ----

# Project schema
rs <- dbGetQuery(pg, "SET search_path TO non_gaap")
cpi <- tbl(pg, "cpi")

# Outside of project schema
funda <- tbl(pg, sql("SELECT * FROM comp.funda"))
fundq <- tbl(pg, sql("SELECT * FROM comp.fundq"))
names <- tbl(pg, sql("SELECT * FROM comp.names"))

# Set params ----

end_date <- '2018-12-31'

# Get perpetual inventory method parameteres ----

library(readxl)
library(googledrive)

# gs_auth()?
# BEA data depreciation rates from 
# Li and Hall (2016) "Depreciation of business R&D capital"
# https://drive.google.com/open?id=1OWhdSF4cki_OavGiEC_ZMnDfdyTd5cD_7HukUqbQ_5Q
file <- drive_download(as_id("1OWhdSF4cki_OavGiEC_ZMnDfdyTd5cD_7HukUqbQ_5Q"), overwrite = TRUE)
delta_bea <- 
    read_excel(file$local_path) %>% 
    mutate(naics = as.integer(naics)) %>%
    select(naics, delta_know_bea = depr_rate) %>% 
    copy_to(pg, ., "delta_bea", temporary = TRUE)

# Ewens, Peters, and Wang (2019) "Acquisition Prices and the Measurement of Intangible Capital"
# https://drive.google.com/open?id=1wOlZLZE9EojTYl0ub2zrjLDz9lZHUSqA6zVDLjNS0wU
file <- drive_download(as_id("1wOlZLZE9EojTYl0ub2zrjLDz9lZHUSqA6zVDLjNS0wU"), overwrite = TRUE)
delta_epw <- 
    read_excel(file$local_path) %>% 
    select(sic, delta_know_epw = knowDepr, gamma_epw = gamma) %>% 
    copy_to(pg, ., "delta_epw", temporary = TRUE)

# Define BEA and Eisfeldt and Papanikolaou (2014) industry assingment ----

# Eisfeldt, Andrea L and Papanikolaou, Dimitris "The value and ownership of intangible capital", AER (2014)
# https://drive.google.com/open?id=1nrIFrGR5aOZ8rhaMpQnREX2dy9tBKvjjhnyRYMzHx6M
# WARNING: "High tech" defined as inidividual SICs that override all other classification
file <- drive_download(as_id("1nrIFrGR5aOZ8rhaMpQnREX2dy9tBKvjjhnyRYMzHx6M"), overwrite = TRUE)
ind_class <- 
    read_excel(file$local_path)

# All SICs
sics <- 
    names %>%
    select(sic) %>%
    mutate(sic = as.integer(sic)) %>%
    union(., funda %>%
              select(sic = sich) %>%
              mutate(sic = as.integer(sic))) %>%
    distinct()

# Fama-French 5 industry-based classification
sics_ff5 <- 
    sics %>%
    mutate(dum = 1) %>%
    distinct() %>%
    # Merge with sic_low != sic_high (non-"High-Tech")
    left_join(ind_class %>%
                  filter(!override) %>%
                  mutate(dum = 1), copy = TRUE) %>%
    filter((sic_low <= sic & sic <= sic_high)) %>%
    select(-override, -sic_low, -sic_high, -dum) %>%
    compute()

# Peters and Taylor (2017) "Intangible capital and the investment-q relation":
# exclude regulated utilities (4900–4999), financial firms (6000–6999), and firms categorized as public
# service, international affairs, or non-operating establishments (9000 and above)

# Eisfeldt and Papanikolaou (2014) industry classification based on SIC
sics_ind <- 
    sics %>%
    left_join(sics_ff5) %>%
    # Regulated utilities (4900–4999)
    mutate(ind_ep = if_else(4900 <= sic & sic <= 4999, "Utilities", ind_ep)) %>%
    # Non-operating establishments
    mutate(ind_ep = if_else(9000 <= sic, "Non-operating est.", ind_ep)) %>%
    # Merge with sic_low == sic_high ("High-Tech") 
    left_join(ind_class %>%
                  filter(override) %>%
                  select(sic = sic_low, ind_ep1 = ind_ep), copy = TRUE) %>%
    # High-tech overrides other industry classification
    mutate(ind_ep = coalesce(ind_ep1, ind_ep)) %>%
    mutate(ind_ep6 = coalesce(ind_ep, "Other")) %>%
    select(-ind_ep, -ind_ep1) %>%
    compute()

# Get quarterly data ----

# WARNING: Acct. values are in millions
colsq <- c("gvkey", "datadate", "fyearq", "fqtr", 
           "xsgaq", # XSGAQ -- Selling, General and Administrative Expenses
           "xrdq", # XRDQ -- Research and Development Expense
           "rdipq", # RDIPQ -- In Process R&D
           "cogsq" # COGSQ -- Cost of Goods Sold
)

# Quarterly fundamentals
fundq_alt <- 
    fundq %>%
    filter(datadate <= end_date) %>%
    filter(indfmt == 'INDL' & datafmt == 'STD' & popsrc == 'D' & consol == 'C' & fic == 'USA') %>%
    select(one_of(colsq)) %>%
    mutate(caldt_year = sql("date_part('year', datadate)"), 
           caldt_month = sql("date_part('month', datadate)")) %>%
    filter(!is.na(datadate) & !is.na(fyearq) & !is.na(fqtr)) %>%
    # For duplicate {gvkey, datadate}, keep min(fyearq)
    group_by(gvkey, datadate) %>%
    filter(fyearq == min(fyearq)) %>%
    ungroup() %>%
    # For remaining duplicate {gvkey, datadate}, keep min(fqtr)
    group_by(gvkey, datadate) %>%
    filter(fqtr == min(fqtr)) %>%
    ungroup() %>%
    compute()

# WARNING: do NOT use year-to-date amount since they close a trivial
# number of missing obs.
fundq_cpi <- 
    fundq_alt %>%
    # Set XSGA, XRD to NA if negative (very(!) few cases)
    mutate_at(.vars = vars(xsgaq, xrdq, cogsq), 
              .funs = funs(if_else(. <= 0, NA, .))) %>%
    # Set missing XSGA and XRD to 0 as in Peters and Taylor (2017)
    # and Eisfeldt and Papanikolaou (2014)
    mutate_at(.vars = vars(xsgaq, xrdq, rdipq, cogsq), 
              .funs = funs(coalesce(., 0))) %>%
    # Extract sga only
    mutate(sgaq = if_else((xrdq > xsgaq & xrdq <= cogsq) | xsgaq == 0, 
                         xsgaq, xsgaq - xrdq - rdipq)) %>%
    # Very(!) few sga < 0, when xrd > xsga & xrd > cogs
    mutate(sgaq = if_else(sgaq < 0, 0, sgaq)) %>%
    # Merge CPI
    left_join(cpi %>%
                  select(caldt_year, caldt_month, cpifac)) %>%
    # Compute real expenses
    mutate_at(.vars = vars(sgaq, xrdq), 
              .funs = funs(cpi = .*cpifac)) %>%
    select(gvkey, datadate, fyearq, fqtr, sgaq_cpi, xrdq_cpi) %>%
    compute()

# Get historical industry membership and annual data ----

# Annual vars
colsa <- c("gvkey", "datadate", "fyear", 
           "fic", # Current ISO Country Code - Incorporation  
           "sich", # SICH -- Standard Industrial Classification - Historical
           "naicsh", # NAICSH -- North America Industrial Classification System - Historical
           "xsga", # XSGA -- Selling, General and Administrative Expense
           "xrd", # XRD -- Research and Development Expense
           "rdip", # RDIP -- In Process R&D Expense
           "cogs" # COGS -- Cost of Goods Sold
           )

# Get historical SIC codes
funda_alt <- 
    funda %>%
    filter(datadate <= end_date) %>%
    filter(indfmt == 'INDL' & datafmt == 'STD' & popsrc == 'D' & consol == 'C' & fic == 'USA') %>%
    filter(!is.na(fyear)) %>% 
    select(one_of(colsa)) %>%
    mutate(caldt_year = sql("date_part('year', datadate)"), 
           caldt_month = sql("date_part('month', datadate)")) %>%
    distinct()

# Industry from names file
names_alt <- 
    names %>% 
    select(gvkey, sic, naics) %>%
    mutate(sic = as.integer(sic),
           naics = as.integer(naics)) %>%
    distinct()

ind_codes <- 
    funda_alt %>%
    # Extrapolate sich backwards and forwards
    mutate(gvkey_partition = sql("sum(CASE WHEN sich IS NULL THEN 0 ELSE 1 END) OVER (PARTITION BY gvkey ORDER BY fyear)")) %>% 
    mutate(gvkey_partition = if_else(is.na(sich), gvkey_partition + 1, gvkey_partition)) %>% 
    mutate(sich = sql("first_value(sich) OVER (PARTITION BY gvkey, gvkey_partition ORDER BY fyear desc)")) %>% 
    arrange(gvkey, fyear) %>% 
    mutate(gvkey_partition = if_else(is.na(sich), gvkey_partition - 1, gvkey_partition)) %>% 
    mutate(sich = sql("first_value(sich) OVER (PARTITION BY gvkey, gvkey_partition ORDER BY fyear)")) %>% 
    select(-gvkey_partition) %>%
    # Fill-in sich, should have no missing sich after this step
    left_join(names_alt) %>% 
    mutate(sich = coalesce(sich, sic)) %>% 
    # Extrapolate naics backwards and forwards
    mutate(gvkey_partition = sql("sum(CASE WHEN naicsh IS NULL THEN 0 ELSE 1 END) OVER (PARTITION BY gvkey ORDER BY fyear)")) %>% 
    mutate(gvkey_partition = if_else(is.na(naicsh), gvkey_partition + 1, gvkey_partition)) %>% 
    mutate(naicsh = sql("first_value(naicsh) OVER (PARTITION BY gvkey, gvkey_partition ORDER BY fyear desc)")) %>% 
    arrange(gvkey, fyear) %>% 
    mutate(gvkey_partition = if_else(is.na(naicsh), gvkey_partition - 1, gvkey_partition)) %>% 
    mutate(naicsh = sql("first_value(naicsh) OVER (PARTITION BY gvkey, gvkey_partition ORDER BY fyear)")) %>% 
    select(-gvkey_partition) %>%
    # Fill-in naicsh, should have no missing naicsh after this step
    left_join(names_alt) %>% 
    mutate(naicsh = coalesce(naicsh, naics)) %>% 
    # Extract first 4 digits from NAICS for BEA data match
    mutate(naicsh4 = sql("substring(naicsh::text, 1, 4)")) %>%
    mutate(naicsh4 = if_else(length(naicsh4) == 4, naicsh4, NA)) %>%
    mutate(naicsh4 = as.numeric(naicsh4)) %>%
    select(gvkey,fyear, sich, naicsh, naicsh4) %>%
    distinct() %>%
    compute()

# Set capital accumulation data ----

# Peters and Taylor (2017), Appendix B1 for SG&A measurement
funda_cpi <- 
    funda_alt %>%
    # Set XSGA, XRD to NA if negative (very(!) few cases)
    mutate_at(.vars = vars(xsga, xrd, cogs), 
             .funs = funs(if_else(. <= 0, NA, .))) %>%
    # Set missing XSGA and XRD to 0 as in Peters and Taylor (2017)
    # and Eisfeldt and Papanikolaou (2014)
    mutate_at(.vars = vars(xsga, xrd, rdip, cogs), 
              .funs = funs(coalesce(., 0))) %>%
    # Extract sga only
    mutate(sga = if_else((xrd > xsga & xrd <= cogs) | xsga == 0, 
                         xsga, xsga - xrd - rdip)) %>%
    # Very(!) few sga < 0, when xrd > xsga & xrd > cogs
    mutate(sga = if_else(sga < 0, 0, sga)) %>%
    # Merge CPI
    left_join(cpi %>%
                  select(caldt_year, caldt_month, cpifac)) %>%
    # Compute real expenses
    mutate_at(.vars = vars(sga, xrd), 
              .funs = funs(cpi = .*cpifac)) %>%
    # Get extrapolated SICs
    select(-sich) %>%
    left_join(ind_codes %>%
                  select(gvkey, fyear, sich)) %>%
    left_join(sics_ind %>%
                  rename(sich = sic)) %>%
    # Get extrapolated NAICS
    select(-naicsh) %>%
    left_join(ind_codes %>%
                  select(gvkey, fyear, naicsh, naicsh4)) %>%
    compute()

# Merge to capital accumulation params
#   (1) BEA assumptions: gamma = 0.3, delta_org = 0.2
#   (2) Ewens et al. (2019) (epw) assumptions from delta_epw data
#   (3) Eisfeldt and Papanikolaou (2013) (ep) assumptions: gamma = 1, delta_org = 0.15, delta_know_ep = 0.15
funda_params <-
    funda_cpi %>%
    # Merge BEA params
    left_join(delta_bea %>%
                  rename(naicsh4 = naics)) %>%
    # WARNING: BEA sets to 0.15 if missing
    mutate(delta_know_bea = coalesce(delta_know_bea, 0.15), 
           gamma_bea = 0.3, 
           delta_org_bea = 0.2) %>%
    # Merge Ewens et al. (2019) params
    left_join(delta_epw %>%
                  rename(sich = sic)) %>%
    # WARNING: BEA sets to 0.15 if missing
    # Set missing gamma_epw to 0.3 (default in the literature)
    mutate(delta_know_epw = coalesce(delta_know_epw, 0.15),
           gamma_epw = coalesce(gamma_epw, 0.3), 
           delta_org_epw = 0.2) %>% 
    # Merge Eisfeldt and Papanikolaou (2013) params
    mutate(delta_know_ep = 0.15, 
           gamma_ep = 1, 
           delta_org_ep = 0.15) %>%
    compute()
    
# Compute initial stock of organization and knowledge capital ----

# Compute SG&A and R&D real growth rates by industry
g_ind <- 
    funda_params %>%
    filter(!is.na(ind_ep6)) %>%
    # Compute annual growth rates
    group_by(gvkey) %>%
    arrange(datadate) %>%
    mutate_at(.vars = vars(sga_cpi, xrd_cpi), 
              .funs = funs(g = (./NULLIF(lag(.), 0) - 1))) %>%
    ungroup() %>%
    select(gvkey, datadate, sich, ind_ep6, matches("_cpi_g")) %>%
    collect() %>%
    # Winsorize at 1/99 by industry
    group_by(ind_ep6) %>%
    mutate_at(.vars = vars(matches("_cpi_g")),
              .funs = funs(winsor(., 0.01))) %>%
    ungroup() %>%
    group_by(ind_ep6) %>%
    summarize_at(.vars = vars(matches("_cpi_g")),
                 .funs = funs(mean(., na.rm = TRUE))) %>%
    ungroup() %>%
    copy_to(pg, ., "g_ind", temporary = TRUE)

# Compute initial capital stock
init_cap <- 
    funda_params %>%
    # Keep the first year by gvkey
    group_by(gvkey) %>%
    filter(datadate == min(datadate)) %>%
    ungroup(gvkey) %>%
    # Merge industry growth
    left_join(g_ind) %>%
    # Compute initial capital stock
    mutate(org_init_bea = gamma_bea*sga_cpi/(sga_cpi_g + delta_org_bea), # BEA asm
           know_init_bea = xrd_cpi/(xrd_cpi_g + delta_know_bea), 
           org_init_epw = gamma_epw*sga_cpi/(sga_cpi_g + delta_org_epw), # EPW asm
           know_init_epw = xrd_cpi/(xrd_cpi_g + delta_know_epw),
           org_init_ep = gamma_ep*sga_cpi/(sga_cpi_g + delta_org_ep), # EP asm
           know_init_ep = xrd_cpi/(xrd_cpi_g + delta_know_ep)) %>%
    compute()

# Compute current stock of organization and knowledge capital ----

calc_cap <- function(cap, invest, delta) {
    # Perpetual inventory method
    for (i in 2:length(cap)){
        cap[i] <- (1 - delta[i])*cap[i - 1] + invest[i]
    }
    return(cap)
}

# Compute current stock using annual data
intanga <- 
    funda_params %>%
    # Drop the first year (in init_cap!)
    group_by(gvkey) %>%
    filter(datadate != min(datadate)) %>%
    ungroup() %>%
    # Initialize current capital stock
    mutate(q_org_bea = 0, q_know_bea = 0,
           q_org_epw = 0, q_know_epw = 0, 
           q_org_ep = 0, q_know_ep = 0, 
           q_org_noinit = 0, q_know_noinit = 0) %>%
    select(gvkey, datadate, fyear, sich, ind_ep6, sga_cpi, xrd_cpi, 
           delta_know_bea, gamma_bea, delta_org_bea, q_org_bea, q_know_bea, 
           delta_know_epw, gamma_epw, delta_org_epw, q_org_epw, q_know_epw,
           delta_know_ep, gamma_ep, delta_org_ep, q_org_ep, q_know_ep, 
           q_org_noinit, q_know_noinit) %>%
    # Add initial capital
    union(., 
          init_cap %>%
              mutate(q_org_noinit = 0, q_know_noinit = 0) %>%
              select(gvkey, datadate,  fyear,sich, ind_ep6, sga_cpi, xrd_cpi, 
                     delta_know_bea, gamma_bea, delta_org_bea, q_org_bea = org_init_bea, q_know_bea = know_init_bea, 
                     delta_know_epw, gamma_epw, delta_org_epw, q_org_epw = org_init_epw, q_know_epw = know_init_epw,
                     delta_know_ep, gamma_ep, delta_org_ep, q_org_ep = org_init_ep, q_know_ep = know_init_ep,
                     q_org_noinit, q_know_noinit)) %>%
    # Compute org and know investment
    mutate(org_bea = gamma_bea*sga_cpi,
           org_epw = gamma_epw*sga_cpi,
           org_ep = gamma_ep*sga_cpi,
           org_noinit = gamma_bea*sga_cpi, 
           know = xrd_cpi) %>%
    # Keep gvkeys with at least 2 obs
    group_by(gvkey) %>%
    filter(n() > 1) %>%
    ungroup() %>%
    collect() %>%
    # Accumulate according to the perpetual inventory method
    group_by(gvkey) %>%
    arrange(datadate) %>%
    # Compute capital stock
    group_modify(~ {
        .x %>%
            # Perpetual inventory method in calc_cap(cap, invest, delta)
            mutate(q_org_bea = calc_cap(q_org_bea, org_bea, delta_org_bea), 
                   q_org_epw = calc_cap(q_org_epw, org_epw, delta_org_epw), 
                   q_org_ep = calc_cap(q_org_ep, org_ep, delta_org_ep), 
                   q_org_noinit = calc_cap(q_org_noinit, org_noinit, delta_org_bea),
                   q_know_bea = calc_cap(q_know_bea, know, delta_know_bea), 
                   q_know_epw = calc_cap(q_know_epw, know, delta_know_epw), 
                   q_know_ep = calc_cap(q_know_ep, know, delta_know_ep), 
                   q_know_noinit = calc_cap(q_know_noinit, know, delta_know_bea)
                   )
        }) %>%
    ungroup() %>%
    copy_to(pg, ., "intanga", temporary = TRUE)

# Compute quarterly capital stock. Use annual data as a starting value
intangq <- 
    fundq_cpi %>%
    left_join(intanga %>%
                  select(gvkey, fyearq = fyear, sich, ind_ep6, 
                         matches("^delta_|^gamma_|^q_")), by = c("gvkey", "fyearq")) %>%
    # Keep only when sich from annual data is NOT missing
    filter(!is.na(sich)) %>%
    # Keep the latest record in the first year
    group_by(gvkey) %>%
    filter(fyearq > min(fyearq) | (fyearq == min(fyearq) & fqtr == max(fqtr))) %>%
    ungroup() %>%
    # WARNING: depreciation rates/4 for quarterly data
    mutate_at(.vars = vars(matches("^delta_")), 
              .funs = funs(./4)) %>%
    # Compute org and know investment
    mutate(org_bea = gamma_bea*sgaq_cpi,
           org_epw = gamma_epw*sgaq_cpi,
           org_ep = gamma_ep*sgaq_cpi,
           org_noinit = gamma_bea*sgaq_cpi, 
           know = xrdq_cpi) %>%
    # Keep gvkeys with at least 2 obs
    group_by(gvkey) %>%
    filter(n() > 1) %>%
    ungroup() %>%
    collect() %>%
    # Accumulate according to the perpetual inventory method
    group_by(gvkey) %>%
    arrange(datadate) %>%
    # Compute capital stock
    group_modify(~ {
        .x %>%
            # Perpetual inventory method in calc_cap(cap, invest, delta)
            mutate(q_org_bea = calc_cap(q_org_bea, org_bea, delta_org_bea), 
                   q_org_epw = calc_cap(q_org_epw, org_epw, delta_org_epw), 
                   q_org_ep = calc_cap(q_org_ep, org_ep, delta_org_ep), 
                   q_org_noinit = calc_cap(q_org_noinit, org_noinit, delta_org_bea),
                   q_know_bea = calc_cap(q_know_bea, know, delta_know_bea), 
                   q_know_epw = calc_cap(q_know_epw, know, delta_know_epw), 
                   q_know_ep = calc_cap(q_know_ep, know, delta_know_ep), 
                   q_know_noinit = calc_cap(q_know_noinit, know, delta_know_bea)
            )
    }) %>%
    ungroup() %>%
    copy_to(pg, ., "intangq1", temporary = TRUE)
    
# Write to PG ----

dbGetQuery(pg, "DROP TABLE IF EXISTS intang")

intangq_alt <- 
    intangq %>%
    # Compute total intangible capital
    mutate(q_bea = q_org_bea + q_know_bea, 
           q_epw = q_org_epw + q_know_epw, 
           q_ep = q_org_ep + q_know_ep,
           q_noinit = q_org_noinit + q_know_noinit, 
           w_bea = org_bea + know, 
           w_epw = org_epw + know, 
           w_ep = org_ep + know, 
           w_noinit = org_noinit + know) %>%
    select(gvkey, datadate, fyearq, fqtr, sich, ind_ep6, 
           matches("^q_|^w_|^org_|delta"), know) %>%
    arrange(gvkey, datadate) %>%
    compute(name = "intang", temporary=FALSE,
            indexes=list(c("gvkey", "datadate"), c("gvkey")))

dbGetQuery(pg, "ALTER TABLE intang OWNER TO non_gaap")
dbGetQuery(pg, "GRANT SELECT ON TABLE intang TO non_gaap_access")
dbGetQuery(pg, "VACUUM ANALYZE intang")

comment <- 'CREATED USING create_intang.R  from GitHub azakolyukina/non_gaap'
db_comment <- paste0("COMMENT ON TABLE intang IS '",
                     comment, " ON ", Sys.time() , "'; ") 
dbGetQuery(pg, db_comment)

dbDisconnect(pg)
