# Initialise array with missing values
@everywhere init_missing(::Type{T}, dims::Integer...) where {T <: Number} = Array{Union{T, Missing}}(missing, dims...);
@everywhere init_missing(dims::Integer...) = init_missing(Float64, dims...);

function winsor_mz(x; prop = 0.0)
    # WARNING: old winsor() function sorted data after winsorization

    x_lb = quantile(x, prop)
    x_ub = quantile(x, 1 - prop)

    # Winsorize at lb/ub
    x[x .< x_lb] .= x_lb
    x[x .> x_ub] .= x_ub
    return x
end

# To initialize policy index, loop over the possibilities in sequential order
function init_pol_old(exog_num, q_num, pol_num)
    pol_old = Array{Int64}(undef, exog_num, q_num)
    ct = 1
    for ii in 1:exog_num, jj in 1:q_num
        if ct > pol_num
            ct = 1
        end
        pol_old[ii, jj] = ct
        ct += 1
    end
    return pol_old
end

function q_grid_pts(q_num, # q_num - number of points in q grid. Must be odd
                    qmid, # qmid - value for midpoint of q grid, i.e., steady-state value
                    delta, # delta - depreciation rate
                    q_space) # spacing between points
    # This function generates grid points for q based on a multiple of 1 - delta around midpoint

    pts_per_side = Int64(q_num/2.0 - 0.5); # number on either side of midpoint
    qgrid = Array{Float64}(undef, q_num); # initilize grid q

    # Insert midpoint into grid
    qgrid[pts_per_side + 1] = qmid

    # Compute grid elemments around mid-point as a multiple of 1 - delta
    for ii in 1:pts_per_side
        qgrid[pts_per_side + 1 - ii] = qgrid[pts_per_side + 1 - ii + 1]*(1.0 - delta)^q_space # pt below
        qgrid[pts_per_side + 1 + ii] = qgrid[pts_per_side + 1 + ii - 1]/(1.0 - delta)^q_space # pt above

    end

    return qgrid
end

function tauchen_calc(rho, # persistence of stochastic process
                      mu, # mean of stochastic process
                      sigma, # standard deviation of stochastic shock z
                      m, # numerical parameter that determines bounds as multiples of std of z
                      zpts); # number of grid point
    # This function calculates transition and vector matrices according to Tauchen (1986) for AR1
    # See Luke Taylor's code from structural camp

    stdev = (m*sigma)/sqrt(1 - rho^2); # range around unconditional mean (assumed 0)
    zmin = mu - stdev; # lower bound
    zmax = mu + stdev; # upper bound

    z = collect(range(zmin, zmax, length = zpts)); # vector of points
    w = (zmax - zmin)/(zpts - 1.0)
    # Calculate transition matrix - ii is index of current value, jj is index for future value
    pr_mat_z = Array{Float64}(undef, zpts, zpts); # pre-allocate transition matrix

    for jj in 1:zpts
        # First point
        z_pt = z[1] - rho*z[jj]
        F1 = cdf(Normal(0,1),(z_pt + w/2.0 - (1.0 - rho)*mu)/sigma)
        pr_mat_z[jj,1] = F1

        # Transition to all but end-points
        for kk in 2:zpts-1
            z_pt = z[kk] - rho*z[jj]
            F1 = cdf(Normal(0,1),(z_pt + w/2.0 - (1.0 - rho)*mu)/sigma)
            F0 = cdf(Normal(0,1),(z_pt - w/2.0 - (1.0 - rho)*mu)/sigma)
            pr_mat_z[jj,kk] = F1 - F0
        end

        # Transition for end point
        z_pt = z[zpts] - rho*z[jj]
        F0 = cdf(Normal(0,1),(z_pt - w/2.0 - (1.0 - rho)*mu)/sigma)
        pr_mat_z[jj,end] = 1.0 - F0
    end

    return z, pr_mat_z
end

function make2Dgrid(xgrid, # 1D grid of x
                    ygrid) # ygrid - 1D grid of y
    # This function reates a 2D grid that includes every combination of xgrid and ygrid
    # Creates 2 matrices, each of dimension length(xgrid)*length(ygrid) x 2
    #          gridval - values of grids. 1st column is x values, 2nd columns is y values
    #          gridind - indices for grids. 1st column is x, 2nd columns is y

    # Compute lengths of each grid
    xnum = length(xgrid)
    ynum = length(ygrid)

    # Initialize outputs
    gridind = Array{Int64}(undef, xnum*ynum, 2)
    gridval = Array{Float64}(undef, xnum*ynum, 2)

    ct = 0; # Initialize row counter

    # Loop through all possible combinations to compute gridval and gridind
    for ii in 1:xnum, jj in 1:ynum
        ct = ct + 1;
        gridind[ct, 1] = ii; # count for w
        gridind[ct, 2] = jj; # count for b

        gridval[ct, 1] = xgrid[ii]; # value for w
        gridval[ct, 2] = ygrid[jj]; # value for b
    end

    return gridval, gridind
end

function q_func(q, alph, delta, kappa_w, theta, kappa_b, mu_pi, r_inv, beta)
    # This function defines implicit equation for q in the steady state for calc_qmid
    # commit 3e3f14a7c35123ce7ebd77af4fb28e31559832bd in non_gaap_drafts
    # WARNING: NEEDS TO BE ADJUSTED FOR (1+r)/r!!!!

    r = (1.0 - r_inv)/r_inv
    # Get beta_1 X non-GAAP earnings, beta_2 x GAAP earnings
    if length(beta) > 2
        beta_1, beta_2 = beta[2:3]
    else
        beta_1 = 0.0
        beta_2 = beta[2]
    end

    lambda = 1/(1 + r) # eq 11
    c = 1 - lambda*(1 - delta) + (1 - lambda*(1 - delta/2))*kappa_w*delta # eq 24
    d = q^alph - (1 + kappa_w/2*delta)*delta*q # eq 28
    if (d > 0)
        q_foc = -c + lambda*alph*q^(alph - 1) + theta*
            (lambda*alph*(beta_1 + beta_2)/2*q^(alph - 1)*d^(-1/2) + theta*lambda*beta_1^2/8/kappa_b*d^(-1) -
            (c*(beta_1 + beta_2)/2 + lambda*mu_pi*beta_2/2)*d^(-1/2)) # eq 31
    else
        q_foc = 1.0 # Arbitrary non-zero value
    end

    return q_foc
end

function calc_qmid(alph, delta, kappa_w, theta, kappa_b, mu_pi, r_inv, beta)
    # This function computes steady-state value for capital

    # Upper bound on root interval at sigma_y = 0.50
    q_bar = (delta/exp(0.125)*(1 + kappa_w/2*delta))^(1/(alph - 1)) # eq 36

    # Find qmid that satisfies steady-state first-order condition
    qmid_flag = 0
    root_val = find_zeros(x -> q_func(x, alph, delta, kappa_w, theta, kappa_b, mu_pi, r_inv, beta), 0.01, q_bar)
    if (length(root_val) > 0)
        # WARNING: max forces d ~ 0
        qmid = minimum(root_val)
    else
        qmid_flag = 1
        qmid  = ((r + delta)/alph)^(1.0/(alph - 1.0)) # Toy investment model from Toni's notes
    end

    # qmid_flag = 1 if steady-state does NOT exists (flags issue with steady_state)
    return qmid, qmid_flag
end

function init_grids(alph, delta, rho_y, sigma_y, sigma_pi, mu_pi, r_inv,
                    theta, beta, kappa_w, kappa_b,
                    q_num, m_y, nu_y_num, m_pi, nu_pi_num, exog_num,
                    b_num, w_num, is_incl_b, is_incl_psi, q_space)

    # Capital grid (q), see issue #127 for the steady-state value of capital
    if length(beta) == 2
        # WARNING: GAAP only case
        beta = [0.0; beta]
    end

    qmid, qmid_flag = calc_qmid(alph, delta, kappa_w, theta, kappa_b, mu_pi, r_inv, beta)
    qgrid = q_grid_pts(q_num, qmid, delta, q_space) # create grid
    minq  = minimum(qgrid) # max value of q on grid
    maxq  = maximum(qgrid) # max value of q on grid
    distq = qgrid[:][2:end] - qgrid[:][1:end-1]

    # Productivity shock (nu_y)
    # WARNING: nu_ygrid corresponds to log(nu_y)
    nu_ygrid, tran_mat_y = tauchen_calc(rho_y, 0.0, sigma_y, m_y, nu_y_num) # compute grid and transition matrix
    minnu_y  = minimum(nu_ygrid) # minimum point on grid
    maxnu_y  = maximum(nu_ygrid) # maximum point on grid

    # Accounting noise (nu_pi)
    nu_pigrid, tran_mat_pi = tauchen_calc(0.0, -mu_pi, sigma_pi, m_pi, nu_pi_num) # compute grid and transition matrix - no autocorrelation
    minnu_pi  = minimum(nu_pigrid) # minimum point on grid
    maxnu_pi  = maximum(nu_pigrid) # maximum point on grid

    # Combine transition matrices from each exogenous state variable
    tran_mat = kron(tran_mat_y, tran_mat_pi) # combined transition matrix

    # Create matrices. 1st dimension is exogenous state variables, 2nd is q
    qmat = repeat(qgrid', exog_num, 1) # current capital

    # Create matrix of values and indices for exogenous variable
    exogval, exogind = make2Dgrid(nu_ygrid, nu_pigrid)

    # Create matrices for exogvals 1 and 2
    exogmat1 = repeat(exogval[:, 1], 1, q_num)
    exogmat2 = repeat(exogval[:, 2], 1, q_num)

    ### Set permissible policy choices ###
    # Create grids and matrices of control variables

    # Create bias grid dense around 0 and sparse in tails
    b_tail = [0.035; 0.05; 0.075; 0.10; 0.125; 0.15] # extreme values of b as a multiple of q
    #b_tail = [];
    b_max = 0.025 # limit on log-grid
    n_steps = Int((b_num - 1.0) - length(b_tail))
    b_step = exp(log(2.0)/n_steps) - 1 # in percentages
    grid_tmp = cumprod(fill(1.0, n_steps) .+ b_step) - fill(1.0, n_steps)
    grid_pos = [b_max*grid_tmp; b_tail]
    bgrid = [0.0; grid_pos]
    bmin = minimum(bgrid); bmax = maximum(bgrid)
    # WARNING: bgrid == sort(bgrid) (must be true)

    # Discretize investment using log-linear spacing
    wmin = 0.01*minq; wmax = 0.35*maxq; # bounds on investment
    wgrid = range(log(wmin), log(wmax), length = w_num)
    wgrid = exp.(wgrid)
    # WARNING: compare wmin/wmax and wgrid[1]/wgrid[w_num]

    # Choice of bias
    bgrid, pol_num = is_incl_b ?
            (bgrid, w_num*b_num) :
            ([0.0], w_num)

    # Set up discretized policy grid, using linear spacing and bounds from above
    pol_val, _ = make2Dgrid(wgrid, bgrid)

    # Initialize policies
    pol_old = init_pol_old(exog_num, q_num, pol_num) # Assign index values to all points
    w = pol_val[pol_old, 1] # initial investment level
    b = pol_val[pol_old, 2] # initial bias in non-GAAP reporting

    # Set initial productivity shock to be average of two exogenous shocks
    initial_shock = repeat((exogval[:, 1] + exogval[:, 2])/2.0, 1, q_num)
    v_old = prod_func(initial_shock, qmat, alph)  # initial value

    return qgrid, minq, maxq, distq, tran_mat,
            qmat,
            exogval, exogind, exogmat1, exogmat2,
            wgrid, wmax, bgrid, bmin, bmax, pol_num, pol_val, pol_old,
            w, b, v_old
end

@everywhere function prod_func(nu_y, # productivity shock (in logs)
                               q, # intangible captial stock
                               alph) # curvature of the prod_func function
    # This function computes the prod_func for the current period
    return exp.(nu_y).*(q.^alph) # lenght(shock) x length(k)
end

@everywhere function expect_div(z, # productivity shock
                                q, # intangible capital stock
                                w, # intangible investment
                                alph, # curvature of the produciton function
                                kappa_w) # adjustment cost of investment
    # This function computes expected cash flow/dividends, d, for the current period
    # Expected dividends: y - w - k/2*w^2
    return prod_func(z, q, alph) .- w .-(kappa_w./2.0).*((w./q).^2.0).*q # length(z) x length(k)
end

@everywhere function non_gaap(b, # bias in non-GAAP adjustment
                              q, # intangible capital stock
                              nu_pi) # accounting noise
    # This function computes non-GAAP adjustment
    return (b .- nu_pi).*q # length(b) x length(nu_pi)
end

@everywhere function weightss(xvec, # 1-D vector of points
                              val, # value to find in along xvec
                              minx, # minimum value of xvec
                              maxx, # maximum value of xvec
                              x_num, # number of points of xvec
                              dist) # vector of distances between points; MUST have same # of elements as xvec
    # This function finds the grid-element weights and indices for interpolation

    # Run loop, without calculating distances to find index of grid point less than val
    ind = 1;
    for _ in 1:x_num
        if (xvec[ind] > val) break end
        ind += 1;
    end
    ind -= 1;

    # Calculate distances, treating corner cases
    if ind in (0, x_num)
        wt = ind/x_num;
        ind += ind == 0 ? 1 : -1;
    else
        wt = (val - xvec[ind])/dist[ind];
    end

    # ind is the index for the grid element just less than val
    # wt is the weight placed on the grid element just GREATER than val
    return ind, wt
end

@everywhere function interp(x, # a scale from which to interpolate
                            b_star, # matrix of bias
                            indq, # index on qgrid from which x was created
                            wtq, # weight on grid point above along q
                            exogpt; #row index for exogenous point
                            near = false) # nearest neighbor
    # This function interpolates x over the state variables

    # Check if one (and only one) point has b == 0
    if ((b_star[exogpt, indq] == 0.0 && b_star[exogpt, indq+1] > 0.0) ||
        (b_star[exogpt, indq] > 0.0  && b_star[exogpt, indq+1] == 0.0)||
        near == true)
        # choose the point closest to the value
        if wtq < 0.5
            return x[exogpt, indq]
        else
            return x[exogpt, indq+1]
        end

    else
        # x_interp - interpolated value of x
        return (1.0 - wtq) * x[exogpt, indq] +  # lower bound of q
                       wtq * x[exogpt, indq+1]  # upper bound of q
    end
end

@everywhere function s_shape(x)
    #return sign.(x).*log.(1 .+ abs.(x))
    return sign.(x).*sqrt.(abs.(x)) # main specification
    #return sign.(x).*abs.(x) # issue #171 linear case
    #return sign.(x).*(abs.(x)).^(3/4) # issue #171 x^(3/4) case
end

@everywhere function curr_utility(q, nu_y, nu_pi, w, b, # grids for states and policies
                                  alph, kappa_w, theta, kappa_b, r_inv, # params
                                  beta, # pricing coef.
                                  is_incl_psi, # whether non-gaap is reported
                                  is_incl_b, # whether bias is permitted
                                  is_incl_w) # whether w is reported
    # WARNING: This function only permits scalar values!!!
    # This function computes current period utility

    # Expected cash flows/dividends
    exp_div = expect_div(nu_y, q, w, alph, kappa_w) # expected dividends

    # GAAP earnings
    gaap = exp_div .+ nu_pi.*q # GAAP earnings

    # Compute current period stock price
    nongaapadj = non_gaap(b, q, nu_pi) # non-GAAP adj.
    nongaap = gaap .+ nongaapadj # non-GAAP earn.
    if is_incl_psi
        if is_incl_b
            prc = beta[1] .+ beta[2].*s_shape.(nongaap) .+ beta[3].*s_shape.(gaap) # pricing
        else
            prc = beta[1] .+ beta[2].*s_shape.(nongaap) # pricing
        end

        b_cost = (kappa_b/2.0).*(b.^2.0).*q # cost from biasing
    else
        prc = beta[1] .+ beta[2].*s_shape.(gaap) # pricing
        b_cost = 0.0 # cost from biasing (set to zero because no non-gaap reporting)
    end

    # Append to pricing function w if is_incl_w = true
    if is_incl_w
        prc = prc .+ beta[end].*w
    end


    # Manager's payoff
    div_m = exp_div .+ theta.*prc .- b_cost

    # dim_m: current-period payoff
    # prc: stock price
    return div_m, prc
end

@everywhere function expect_pay(q, # amount of q at that point
                                jj, # index along exogenous vector
                                tran_mat, # 2D transition matrix
                                exog_num, # number of points on grid along exogenous dimension
                                v, # 2D matrix of value. 1st dim is exognenous, 2nd is q
                                w, # scalar of investment
                                delta, # depreciation rate
                                qgrid, # grid of q values
                                minq, maxq, # min/max value of q
                                q_num, # number of points on grid along q dimension
                                distq) # distance between points
    # This function computes the expected future payoff for a single point

    # Compute future value of q
    q_new = (1.0 - delta)*q + w; # q_new
    indq, wtq = weightss(qgrid, q_new, minq, maxq, q_num, distq); # index and weights for q_new

    v_single_point = 0.0; # initialize E(v)

    # Compute future value for each value of next period's exogenous value
    for kk in 1:exog_num
        @inbounds v_single_point +=
            tran_mat[jj, kk]*((1.0 - wtq)*v[kk, indq] + wtq*v[kk, indq + 1])
    end

    # E(v) using the law-of-motion for q and transition matrix for exogenous variables
    return v_single_point
end

@everywhere function howard_accel(tran_mat, # {nu_y_num x nu_pi_num} transition matrix
                                  exog_num, # number of points on grid along exogenous dimension
                                  exogval, # {(nu_y_num*nu_pi_num) x 2} matrix of exogenous values
                                  v, # {(nu_y_num*nu_pi_num) x q_num} matrix of value. 1st dim is exognenous, 2nd is q
                                  w, # {(nu_y_num*nu_pi_num) x q_num}  matrix of investment. 1st dim is exognenous, 2nd is q
                                  b, # {(nu_y_num*nu_pi_num) x q_num}  matrix of bias. 1st dim is exognenous, 2nd is q
                                  alph, # curvature of the prod_func function
                                  delta, # depreciation rate
                                  kappa_w, # adjustment costs to investment
                                  theta, # relative weight on the stock price in the manager's obj.
                                  kappa_b, # personal cost from bias
                                  r_inv, # discount rate
                                  beta, # pricing coeff
                                  qmat, # q capital matrix
                                  qgrid, # grid of q values
                                  minq, maxq, # min/max value of q
                                  q_num, # number of points on grid along q dimension
                                  distq, # distance between points in qgrid
                                  exogmat1, # nu_y shock to productivity
                                  exogmat2, # nu_pi shock to GAAP earnings
                                  policy_tol, # distance between iterations for value function
                                  howard_iter, # maximum number of Howard accelerations
                                  is_incl_psi, # whether non-gaap is reported
                                  is_incl_b, # whether bias is permitted
                                  is_incl_w) # whether w is reported
    # This function performs Howard accelerations and outputs value funtion

    # Initialize distance and iterations
    dist = Inf # distance
    iter = 0 # iterations

    # Initialize current value and pricing
    curr_val = Array{Float64}(undef, exog_num, q_num)
    prc = Array{Float64}(undef, exog_num, q_num)

    # Determine current value and pricing
    for ii in 1:q_num
        for jj in 1:exog_num
            curr_val[jj, ii], prc[jj, ii] = curr_utility(qmat[jj,ii], exogmat1[jj,ii], exogmat2[jj,ii],
                                                        w[jj,ii], b[jj,ii],
                                                        alph, kappa_w, theta, kappa_b,
                                                        r_inv, beta, is_incl_psi, is_incl_b, is_incl_w)
            # Fix prc < 0 observations to be a small value
            if prc[jj, ii] < 0.0
                prc[jj, ii] = 0.0001
            end

        end
    end

    while dist > policy_tol && iter < howard_iter

        # Initialize new value function
        v_new = Array{Float64}(undef, exog_num, q_num)

        # Compute the value function for each point
        for ii in 1:q_num
            for jj in 1:exog_num

                @inbounds v_new[jj, ii] =
                    curr_val[jj, ii] +
                    r_inv*expect_pay(qgrid[ii], jj, tran_mat, exog_num, v,
                                          w[jj, ii], delta, qgrid, minq, maxq, q_num, distq); # future payoff
            end
        end

        # WARNING: mask non-positive prc
        for ii in 1:size(prc, 1)
            for jj in 1:size(prc, 2)
                if (prc[ii, jj] <= 0)
                    v_new[ii, jj] = -Inf
                end
            end
        end

        # Compute distance between iterations
        dist = maximum(abs.(v_new - v));

        # Update value function and iteration count
        v = v_new;
        iter += 1;
    end

    # v: value function exog_num x q_num
    return v
end

@everywhere function optimal_policy(ii, # ii - index for qgrid
                                    tran_mat, exog_num, exogval, v, pol_num, pol_val, # transition matrix and states
                                    alph, delta, kappa_w, theta, kappa_b, r_inv, # params
                                    beta, # price coef.
                                    qgrid, minq, maxq, q_num, distq, is_incl_psi, is_incl_b, is_incl_w)
    # This function computes the optimal policy for a given continuation payoff
    q = qgrid[ii] # q for selected grid point
    polmat = Array{Int64}(undef, exog_num, 1) # intialized policies

    for jj in 1:exog_num

        # Initialize value amounts
        v_options = Array{Float64}(undef, pol_num) # initialize vector of policies

        for ct in 1:pol_num

            # Extract consideration policies
            w = pol_val[ct, 1] # investment
            b = pol_val[ct, 2] # bias

            # Compute value function for choice of w and b
            curr_val, prc =
                curr_utility(q, exogval[jj, 1], exogval[jj, 2],
                             w, b, alph, kappa_w, theta, kappa_b, r_inv, beta, is_incl_psi, is_incl_b, is_incl_w)
                 # WARNING: mask non-positive prc
                 if prc > 0
                @inbounds v_options[ct] = curr_val +
                        r_inv*expect_pay(q, jj, tran_mat, exog_num, v, w,
                                        delta, qgrid, minq, maxq, q_num, distq) # future payoff
                else
                # Value if prc is negative
                    v_options[ct] = -Inf
                end

        end

        v_max, pol_ind_optim = findmax(v_options) # find index of controls that max v
        polmat[jj, 1] = pol_ind_optim # return column vector of optimal controls
    end
    # A matrix of the locations (indices) of the optimal policies
    return polmat
end

@everywhere function fund_value(tran_mat, # 2D transition matrix
                                exog_num, # number of points on grid along exogenous dimension
                                exogval, # 2D matrix of exogenous values
                                w, # 2D matrix of investment. 1st dim is exognenous, 2nd is q
                                alph, # curvature of the prod_func function
                                delta, # depreciation rate
                                kappa_w, # adjustment cost of investment
                                r_inv, # r - discount rate
                                qgrid, # grid of q values
                                minq, maxq, # min/max value of q
                                q_num, # number of points on grid along q dimension
                                distq, # distance between points
                                howard_iter, # maximum number of howard accelerations
                                policy_tol) # distance between iterations for value function
    # This function computes the fundamental value of the firm

    # Initialize iteration counter and distance
    dist = Inf
    iter = 0

    # Initialize fund. value as current expected dividends
    expect_div_mat = Array{Float64}(undef, exog_num, q_num)
    for ii in 1:q_num
        for jj in 1:exog_num
            @inbounds expect_div_mat[jj, ii] = expect_div(exogval[jj, 1], qgrid[ii], w[jj, ii], alph, kappa_w)
        end
    end
    vf  = expect_div_mat

    # Loop through pricing function until it converges
    while iter < howard_iter && dist > policy_tol

        vf_new = Array{Float64}(undef, exog_num, q_num)

        # Compute the value function for each point
        for ii in 1:q_num
            for jj in 1:exog_num

                # Insert current period dividends into fundamental value
                @inbounds vf_new[jj, ii] = expect_div_mat[jj, ii] + r_inv*expect_pay(qgrid[ii], jj, tran_mat, exog_num, vf, w[jj, ii],
                                                              delta, qgrid, minq, maxq, q_num, distq)
            end
        end

        dist = maximum(abs.(vf - vf_new)) # distance between iterations
        iter += 1 # update count
        vf = vf_new # update fundamental value
    end

    return vf
end

function compute_betas(x, # array of dependent variables
                    vfvec) # vector of values
    # This function checks that the regression is is non-singular
    sing_flag = 0 # set singular flag to 0

    # Check if x is signular set sing_flag to 1 and exit
    try
        det_x = inv(x'*x)
    catch
        coef = missing #set coef to missing as the singular flag will exit loops
        sing_flag = 1 # set singular flag to 1
    end

    if sing_flag != 1 # Continue if no sing_flag
        # Compute fully rational expectation
        coef = inv(x'*x)*x'*vfvec # compute coefficients using OLS
    end

    return coef, sing_flag
end

function signal_weights(tran_mat, exog_num, exogval, w, b,
                        alph, delta, kappa_w, r_inv,
                        min_pts, qgrid, minq, maxq, q_num, distq,
                        howard_iter, policy_tol, fixed_coef, is_incl_psi, is_incl_b, is_incl_w)

    # Initialize vectors for regressions - nongaap
    gaap    = Array{Float64}(undef, exog_num, q_num) # GAAP earnings matrix
    nongaapadj = Array{Float64}(undef, exog_num, q_num) # non-GAAP adj. matrix
    sing_flag = 0 # Singular matrix  flag

    # Compute GAAP and non-GAAP signals
    for ii in 1:q_num
        for jj in 1:exog_num
            @inbounds gaap[jj, ii] = expect_div(exogval[jj, 1], qgrid[ii], w[jj, ii], alph, kappa_w) + qgrid[ii]*exogval[jj, 2] # GAAP earnings
            @inbounds nongaapadj[jj, ii] = non_gaap(b[jj, ii], qgrid[ii], exogval[jj, 2]) # non-GAAP adj.
        end
    end

    # Compute fundamental value
    vf = fund_value(tran_mat, exog_num, exogval, w,
                    alph, delta, kappa_w, r_inv,
                    qgrid, minq, maxq, q_num, distq,
                    howard_iter, policy_tol)

    # Convert matrices into vectors
    vfvec = vec(vf) # fundamental values

    # Check that all values are greater than 0
    vf_flag = minimum(vfvec) < 0.0 ? 1 : 0

    # Convert matrices into vectors
    onevec = ones(q_num*exog_num);# intercept
    gaapvec = vec(gaap) # GAAP earn.
    nongaapadjvec = vec(nongaapadj) # non-GAAP adj.
    nongaapvec = gaapvec .+ nongaapadjvec # non-GAAP earn.
    gaapvecss = s_shape.(gaapvec) # GAAP earn. S-shaped
    nongaapvecss = s_shape.(nongaapvec) # non-GAAP earn. S-shaped
    if is_incl_w # create a vector of w only if is_incl_w = true
        wvec = vec(w) # w
    end
    coeffs = nothing

    if is_incl_psi
      # Fully rational expectation with non-GAAP signal
      # Combine signals (and intercept) vectors into a n x 4 matrix
      if is_incl_b
          x = [onevec nongaapvecss gaapvecss]
      else
          x = [onevec nongaapvecss]
      end
    else
      # Fully rational expectation without non-GAAP signal
      # Combine signals (and intercept) vectors into a Nx4 matrix
      x = [onevec gaapvecss]
    end
    if is_incl_w
        x = [x wvec]
    end

    # Check if x is signular set sing_flag to 1 and exit
    try
      det_x = inv(x'*x)
    catch
      coeffs = missing #set coef to missing as the singular flag will exit loops
      sing_flag = 1 # set singular flag to 1
    end

    if sing_flag != 1 # Continue if no sing_flag
      # Compute fully rational expectation
      coeffs = inv(x'*x)*x'*vfvec # compute coefficients using OLS
    end

    return coeffs, sing_flag, vf_flag, vf, vfvec, nongaapvec, gaapvec # return outputs
end

function rational_expect(tran_mat, exog_num, exogval, v_old,
                         pol_old, pol_num, pol_val,
                         w, b,
                         alph, delta, kappa_w, theta, kappa_b, beta_old, coeff_damp,
                         r_inv, qmat, qgrid, minq, maxq, q_num, distq, exogmat1, exogmat2,
                         policy_tol, price_tol, howard_iter, max_iter, price_iter, fixed_coef,
                         is_incl_psi, is_incl_b, is_incl_w)
    # This function computes the optimal policies given rational expectations

    # Initialize iteration count and distance
    iter = 0
    dist = Inf # Relative distance on coefficients
    abs_dist = Inf # Absolute distance on coefficients
    vf_flag = 3 # Initial vf_flag
    sing_flag = 3 # Initial coefficient flag

    v = v_old # Value to the manager
    vf = v_old # Fundamental firm value

    vfvec = Array{Float64}(undef, exog_num*q_num)
    nongaapvec = Array{Float64}(undef, exog_num*q_num)
    gaapvec = Array{Float64}(undef, exog_num*q_num)

    # Loop through rational expectation
    while (iter < price_iter) && (dist > price_tol && abs_dist > 1e-4)

        # Compute optimal policies
        w_star, b_star, v = model(tran_mat, exog_num, exogval, v,
                                       pol_old, pol_num, pol_val,
                                       w, b,
                                       alph, delta, kappa_w, theta, kappa_b, r_inv,
                                       beta_old,
                                       qmat, qgrid, minq, maxq, q_num, distq,
                                       exogmat1, exogmat2,
                                       policy_tol, howard_iter, max_iter, is_incl_psi, is_incl_b, is_incl_w)

        # Compute pricing coefficient
        beta_new, sing_flag, vf_flag, vf, vfvec, nongaapvec, gaapvec =
            signal_weights(tran_mat, exog_num, exogval, w_star, b_star,
                           alph, delta, kappa_w, r_inv, min_pts, qgrid, minq, maxq, q_num, distq,
                           howard_iter, policy_tol, fixed_coef, is_incl_psi, is_incl_b, is_incl_w)

        # Exit loop if matrix in regression is singular
        if (sing_flag == 1)
            w = w_star;
            b = b_star;
            if is_incl_psi
                beta_old = [-99.0; -99.0; -99.0]
            else
                beta_old = [-99.0; -99.0]
            end
            priceflag = 0
            return w, b, beta_old, priceflag, sing_flag, vf_flag, vf, vfvec, nongaapvec, gaapvec
        end

        # Update pricing coefficients - compute %diff and update iter ct
        dist = maximum(abs.(beta_new./beta_old .- 1.0)); # distance for each parameter
        abs_dist = maximum(abs.(beta_new - beta_old))
        beta_old = (1.0 - coeff_damp).*beta_old .+ coeff_damp.*beta_new; # dampened coefficients

        iter += 1;
        w = w_star;
        b = b_star;
    end

    # Create a priceflag to indicate RE exceeded iteration count
    priceflag = Int(iter == price_iter) # set to 0

    # Print pricing coefficients
    println("Price coeffs = ", beta_old);

    # w - matrix of optimal investment
    # b - matrix of optimal manipulation
    # beta_old - vector of linear pricing coefficients
    # vf firm value
    # v value to the manager
    return w, b, beta_old, priceflag, sing_flag, vf_flag, vf, v, vfvec, nongaapvec, gaapvec
end

function ar1(xvec, # vector of data to compute the AR1
             obs) # obs - number of firms in the data
    # This function computes an AR1 regression

    # Create a lagged vector. When julia creates vectors, it orders column 1, column 2 etc.
    xvec_lag = xvec[1:end-obs]; # drop last period
    xvec = xvec[obs+1:end]

    # Compute AR coefficients
    one_vec = ones(length(xvec)) # intercept vector
    x = [one_vec xvec_lag] # combine intercept with lagged variable

    b = try inv(x'*x)*x'*xvec catch err; [-99.0; -99.0] end

    # A scalar with the intercept from the AR(1) regression
    return b[2]
end

function reg_lag(yvec, # vector of dependent variables
                xvec, # vector of covariates
                obs) # obs - number of firms in the data
    # This function computes a regression of lagged x's on current y's

    # Create a lagged vector. When julia creates vectors, it orders column 1, column 2 etc.
    xvec_lag = xvec[1:end-obs]; # drop last period
    yvec = yvec[obs+1:end]

    # Compute AR coefficients
    one_vec = ones(length(xvec_lag)) # intercept vector
    x = [one_vec xvec_lag] # combine intercept with lagged variable
    b = inv(x'*x)*x'*yvec # compute coefficients

    # A scalar with the intercept from the AR(1) regression
    return b[2]
end


function backspaces(num)
    ret = num < 0 ? "\b" : ""
    for _ in 1:Int(floor(log10(abs(num))))
        ret = string(ret, "\b")
    end
    #println(@sprintf "Backspaces for %.8f: 321%s\t%sOi with length %i, %i" num ret ret length(ret) Int(floor(log10(abs(num)))))
    return ret
end

spaces(num) = (num >= 0) ? " " : ""

# Macro that runs expression provided whilst suppressing print & println calls
# Doesn't work in nested function calls bc `quote' expression is evaluated in global scope...
macro noprint(expr)
    quote
        function temp_noprint_func()
            println(x...) = nothing;
            print(x...) = nothing;
            return $expr
        end
        temp_noprint_func()
    end
end

function mean_cond(x, dum)
    # Conditional mean
    # E(x | dum = 1), dum = {0, 1}
    res = mean(x.*dum)./mean(dum)

    return res
end

function cov_cond(x, y, dum)
    # Conditional covariance
    # cov(x, y | dum = 1), dum = {0, 1}
    res = mean(x.*y.*dum)./mean(dum) .- (mean(x.*dum)./mean(dum)).*(mean(y.*dum)./mean(dum))
    return res
end


function skew_cond(x, dum)
    # Conditional skewness (see note_on_skewness)
    # E[(x - E[x])^3 | dum = 1]/E[(x - E[x])^2 | dum = 1]^3/2, dum = {0, 1}
    mu_cond_x3_dum = mean((x.^3.0).*dum)./mean(dum)
    mu_cond_x2_dum = mean((x.^2.0).*dum)./mean(dum)
    mu_cond_x_dum = mean(x.*dum)./mean(dum)
    F_dum = mu_cond_x3_dum .- 3.0.*mu_cond_x2_dum.*mu_cond_x_dum .+ 2.0.*mu_cond_x_dum.^3.0
    G_dum = mu_cond_x2_dum .- mu_cond_x_dum.^2.0
    res = F_dum./(G_dum.^(1.5))
    return res
end


function counter_mean(x, # vector of parameters
                        w_star, # matrix of optimal choices
                        b_star, # matrix of optimal bias
                        vf, # matrix of fundamental value
                        sim_data, # simulated data from base case
                        qgrid, # q grid
                        exogval, # 2 column vector of exogenous values
                        beta, # pricing coefficients
                        wins; # level of winsorization
                        base = false, # whether or not its the base case
                        is_fb = false, # whether or not this is the first-best
                        is_incl_psi = true, # whether we include psi in pricing function
                        is_incl_b = true, # whether we include bias (if not, only price off of non-gaap)
                        is_incl_w = false, # whether we include w in pricing
                        return_data = false) # whether to return data or just means
    # This function computes the mean amounts for variables counterfactual metrics
    # Extract parameter values
    alph, kappa_w, rho_y, sigma_y, sigma_pi, theta, kappa_b, mu_pi = x
    kappa_b = kappa_b*10.0 # WARNING: change kappa_b units

    # This function interpolates policy function and vf for each counterfactual
    qsim = sim_data[:,:q] # Create a vector q variables
    nu_ysim = sim_data[:,:nu_y] # Create a vector nu_y variables
    nu_pisim = sim_data[:,:nu_pi] # Create a vector nu_pi variables
    maxq = maximum(qgrid) # maximum value of qgrid
    minq = minimum(qgrid) # minimum value of qgrid
    distq = qgrid[:][2:end] - qgrid[:][1:end-1] # distance between points

    # Interpolate points if !base case. If base case, then take simulated data
    if base
        wsim = sim_data[:,:w]
        bsim = sim_data[:,:b]
        vfsim = sim_data[:,:vf]
    else
        # Initialize vectors of policies and vf
        num_obs = length(qsim)
        wsim = Array{Float64}(undef, num_obs, 1) # w
        bsim = Array{Float64}(undef, num_obs, 1) # b
        vfsim = Array{Float64}(undef, num_obs, 1) # b

        for ii in 1:num_obs
            # Location of observation in q state space
            indq, wtq = weightss(qgrid, qsim[ii], minq, maxq, q_num, distq)

            # Find Location of observation in exogenous parameters
            idx = 1
            while nu_ysim[ii] != exogval[idx, 1] || nu_pisim[ii] != exogval[idx,2]
                idx += 1
            end

            # Optimal choices and vf
            wsim[ii]  = interp(w_star, b_star, indq, wtq, idx)
            bsim[ii]  = interp(b_star, b_star, indq, wtq, idx)
            vfsim[ii] = interp(vf, b_star, indq, wtq, idx)
        end
    end

    # Compute additional variables
    dsim = expect_div(nu_ysim, qsim, wsim, alph, kappa_w)
    gaapsim = dsim .+ qsim.*nu_pisim # gaap
    non_gaapadjsim = non_gaap(bsim, qsim, nu_pisim) # non-gaap adj.
    nongaapsim = gaapsim .+ non_gaapadjsim # non-GAAP earn.
    w_piw = 100.0*wsim./(wsim .+ kappa_w.*0.5*((wsim./qsim).^2.0).*qsim .+ gaapsim)
    w_dw = 100.0*wsim./(wsim .+ kappa_w.*0.5*((wsim./qsim).^2.0).*qsim .+ expect_div(nu_ysim, qsim, wsim, alph, kappa_w))
    w_q = 100.0*wsim./qsim
    b_bnupi = 100.0.*bsim.*qsim./(bsim.*qsim .+ abs.(non_gaapadjsim))

    # Compute price if not first-best
    if !is_fb
        if is_incl_psi
            if is_incl_b
                prc = beta[1] .+ beta[2].*s_shape.(nongaapsim) .+ beta[3].*s_shape.(gaapsim) # pricing
            else
                prc = beta[1] .+ beta[2].*s_shape.(nongaapsim) # pricing
            end
        else
            prc = beta[1] .+ beta[2].*s_shape.(gaapsim) # pricing
        end
        if is_incl_w
            prc = prc .+ beta[end].*wsim
        end
    end

    # Winsorize Data
    wsim = collect(winsor_mz(wsim[:,1], prop = wins))
    dsim = collect(winsor_mz(dsim[:,1], prop = wins))
    gaapsim = collect(winsor_mz(gaapsim[:,1], prop = wins))
    non_gaapadjsim = collect(winsor_mz(non_gaapadjsim[:,1], prop = wins))
    nongaapsim = collect(winsor_mz(nongaapsim[:,1], prop = wins))
    w_piw = collect(winsor_mz(w_piw[:,1], prop = wins))
    w_dw = collect(winsor_mz(w_dw[:,1], prop = wins))
    w_q = collect(winsor_mz(w_q[:,1], prop = wins))
    b_bnupi = collect(winsor_mz(b_bnupi[:,1], prop = wins))

    # Determine residual uncertainty as a regression of vf = a + b*prc + err
    errs = 0 # create a placeholder for error
    if !is_fb
        prc = collect(winsor_mz(prc[:,1], prop = wins)) # winsorize price
        vfsim = collect(winsor_mz(vfsim[:,1], prop = wins)) # winsorize vf
        one_vec = ones(length(prc)) # intercept vector
        x = [one_vec prc] # Combine prc and ones vector
        coef = inv(x'*x)*x'*vfsim # Estimate coefficients
        errs = vfsim .- x*coef # compute error term
    end

    # Determine whether to return data or not
    if !return_data
        # Compute summary statistics for counterfactuals
        return [mean(b_bnupi), mean(wsim), mean(gaapsim), mean(dsim), mean(w_piw), mean(w_dw), mean(w_q), mean(vfsim), sqrt(var(errs)./var(vfsim))]
    else
        temp_mat = hcat(sim_data[:,:q], wsim, bsim, nu_ysim, nu_pisim)
        sim_data_counter = DataFrame(temp_mat)
        rename!(sim_data_counter, [:q, :w, :b, :nu_y, :nu_pi])

        return sim_data_counter
    end
end
