# This file defines all settings for the model

# WARNING: path here should point to non_gaap/model

using Random, DelimitedFiles

# Define params of the model

# Initialize setup (in case setup.txt fails)
suffix = "bea" # Intangibles type
is_incl_b = true # Include bias into the model (main spec)
is_incl_w = false # Include w in pricing
is_price_fixed = false # Fix price from the data---NOT rational expectations

# Replace by setup.txt
setup = readdlm(string(path, "/Output/setup.txt"))
suffix = setup[1] # intangible capital specification
is_incl_b = setup[2] # include bias into the model (main spec)
is_incl_w = setup[3] # Include w in pricing
is_price_fixed = setup[4] # fix price from the data---NOT rational expectations

# Define params outside of setup---mostly for counterfactuals
is_incl_psi = true # Set to false if we don't want non-gaap reporting
is_comp_static = false # Set to false so we dont return only model solution
is_x_grid_x = true # Use the same parameters for grids as those used in the model
x_grid = [] # Use the same parameters for grids as those used in the model
price_coef_grid = [] # Price coefficients for grids. Set to be those estimated in the model

# Read in parameter range
lb_ub = readdlm(string(path, "/Output/lb_ub.txt"), ';', Float64)

#-------------------------------------------------------------------------------------------------------#
# Load data moments, weight matrix, delta parameter, and shuffle parameter

moments  = readdlm(string(path, "/../moments_weights/csv/mom_", suffix, ".csv"), ','; skipstart = 1)
data_moments = convert(Vector{Float64}, moments[:,2]) # extract column of moments

# Load CSV of weight matrix
Qmm = readdlm(string(path, "/../moments_weights/csv/Qmm_", suffix, ".csv"), ','; skipstart = 1)
Qmm = convert(Matrix{Float64}, Qmm[:, 2:end]) # covariance matrix of moments

# Optimal clustered matrix
Omega = readdlm(string(path, "/../moments_weights/csv/Qmm_opt_clust_firm_", suffix, ".csv"), ','; skipstart = 1)
Omega = convert(Matrix{Float64}, Omega[:, 2:end])

# Load csv with shuffle parameters
pshuffle  = readdlm(string(path, "/../moments_weights/csv/pshuffle_", suffix, ".csv"), ','; skipstart = 1)
pshuffle = pshuffle[1, 2] # convert to float matrix

# Load csv with delta paramter---no column name
delta  = readdlm(string(path, "/../moments_weights/csv/delta_", suffix, ".csv"), ',')
delta  = delta[1, 1] # delta parameter

# Load csv with pricing parameters
# WARNING: use full-sample estimates ONLY at price_coef_bea
# (smaller samples, e.g., bea_mb2, can have a POSITIVE weight on psi2)
price_coef  = readdlm(string(path, "/../moments_weights/csv/price_coef_bea.csv"), ','; skipstart = 1)
price_coef = convert(Matrix{Float64}, price_coef[:, 2:end]) # convert to float matrix

# Set fixed coefficients
fixed_coef = [0.0]
# {1, s_shape(pi + psi), s_shape(pi)}
price_coef_init = [0.0; price_coef[2]; price_coef[3]] # inital guess for price coefficients for RE loop
#price_coef_init = [0.0; 10.0; 10.0] # inital guess for price coefficients for RE loop [1, g(pi + psi), g(pi)]
if is_incl_w
    price_coef_init = [price_coef_init; price_coef[3]] # initial guess for price coefficients for RE if including w. !Must be non-zero!
end

#-------------------------------------------------------------------------------------------------------#
# Define parameters that are assumed

r = 0.015; #discount rate - assumed
r_inv = 1.0/(1.0 + r)

#-------------------------------------------------------------------------------------------------------#
# Set tolerances, number of grid points, etc.

# Set variables related to state space
q_num     = 21; # {21} number of points for q. !!!Must be odd!!!
q_space   = 3.0; # The spacing between points of q (original q_space = 2.0!)
nu_y_num  = 9; # number of points for nu_pi state
nu_pi_num = 9; # number of points for nu_pi state
exog_num  = nu_y_num*nu_pi_num; # number of grid points for exogenous parameter
m_y      = 3.0; # number of standard deviations to incorporate in grid
m_pi     = 3.0; # number of standard deviations to incorporate in grid

# Set variables associated with choice variables
w_num   = 61; # number of points for investment policy
b_num   = 61; # number of points for bias policy

# Minimum number of points needed for each pricing scenarion
min_pts = 10;

# Set price-iter. coeffs.
coeff_damp = 0.10; # {0.05, 0.10, 0.3} dampening parameter on coefficients

# Set variables associated with simulations
sims  = 20 # number of simulations
obs   = 1500 # extract number of firms
tburn = 200; # number of periods for burn-in period
t     = 40 + tburn + 1; # number of periods (adjusting for burn-in and last period drop)
obst  = obs*t; # combination of observations and time period
wins  = 0.05 # level of winsorization for simulated data before computing moments (!)

Random.seed!(2019) # fix seed
qseed         = rand(obs, 1, sims); # set iniital value for q - will get adjusted for bounds in function
exog_uniform  = rand(obs, t, sims); # uniform draws equal to size of simulated data
shuffleshocks = rand(obs, t, sims); # random shocks to cash flow shuffling

# Set tolerances
policy_tol       = 1e-7; # inner loop of policy iteration
policy_tol_outer = 1e-6; # outer loop of policy iteration
price_tol        = 0.05; # pricing tolerance as a percent change - based on Terry et al. (2018)
max_iter         = 25; # maximum number of iterations for policy iteration
howard_iter      = 500; # number of Howard accelerations
price_iter       = 150; # number of price iterations
bunchtol         = 0.05; # {5e-2} bunching tolerance on grids
max_sim_bunch    = 0; # maxmimum number of simulations that can be bunched

#-------------------------------------------------------------------------------------------------------#
# Define moments

# Set threshold for psi_q conditioning, i.e,. psi > psi_q_lb
psi_q_lb = 0.0001

# WARNING: set moment index to use
# see calc_moments: sim_moment[ss, :]...
# pi is GAAP earnings; d is dividiends; psi is non-GAAP adj.
mom_labels =["E(w/q)", # 1   | mom_wq
             "E(pi/q)", # 2   | mom_piq
             "rho(pi/q)", # 3   | mom_rho_piq
             "E(d/q)", # 4   | mom_dq
             "rho(d/q)", # 5   | mom_rho_dq
             "E(psi/q)", # 6   | mom_psiq
             "rho(psi/q)", # 7   | mom_rho_psiq
             "Cov(Dw, Dw)", # 8   | mom_w_w
             "Cov(Dw, Dpi)", # 9   | mom_w_pi
             "Cov(Dw, Dd)", # 10  | mom_w_d
             "Cov(Dw, Dpsi)", # 11  | mom_w_psi
             "Cov(Dpi, Dpi)", # 12  | mom_pi_pi
             "Cov(Dpi, Dd)", # 13  | mom_pi_d
             "Cov(Dpi, Dpsi)", # 14  | mom_pi_psi
             "Cov(Dd, Dd)", # 15  | mom_d_d
             "Cov(Dd, Dpsi)", # 16  | mom_d_psi
             "Cov(Dpsi, Dpsi)", # 17  | mom_psi_psi
             "rho(pi, ping_lag)", # 18  | mom_rho_pipingq
             "Cov(DDprc, Dpi)", # 19  | mom_dvf_pi
             "Cov(DDprc, Dpsi)", # 20  | mom_dvf_psi
             "Cov(DDprc, Dpi2)", # 21  | mom_dvf_pi2
             "Cov(DDprc, Dpsi2)", # 22  | mom_dvf_psi2
             "Cov(DDprc, Dpipsi)", # 23  | mom_dvf_pipsi
             "Cov(DDprc, Dping)", # 24  | mom_dvf_ping
             "E(psipos)", # 25  | mom_psipos
             "E(psi/q | psipos)", # 26  | mom_psiq_psipos
             "Cov(Dw, Dw | psipos)", # 27  | mom_w_w_psipos
             "Cov(Dw, Dpi | psipos)", # 28  | mom_w_pi_psipos
             "Cov(Dw, Dd | psipos)", # 29  | mom_w_d_psipos
             "Cov(Dw, Dpsi | psipos)", # 30  | mom_w_psi_psipos
             "Cov(Dpi, Dpi | psipos)", # 31  | mom_pi_pi_psipos
             "Cov(Dpi, Dd | psipos)", # 32  | mom_pi_d_psipos
             "Cov(Dpi, Dpsi | psipos)", # 33  | mom_pi_psi_psipos
             "Cov(Dd, Dd | psipos)", # 34  | mom_d_d_psipos
             "Cov(Dd, Dpsi | psipos)", # 35  | mom_d_psi_psipos
             "Cov(Dpsi, Dpsi | psipos)", # 36  | mom_psi_psi_psipos
             "Cov(DDvf, Dpi | psipos)", # 37  | mom_dvf_pi_psipos
             "Cov(DDvf, Dping | psipos)", # 38  | mom_dvf_ping_psipos
             "Skew(psi/q | psipos)", # 39  | mom_skew_psiq_psipos
             "Beta_pipsi", # 40  | mom_beta_pipsi
             "Beta_pi", # 41  | mom_beta_pi
             "E(psizero)", # 42  | mom_psizero
             "E(pi/q | psipos)" # 43  | mom_piq_psipos
             ]
mom_index = [1:2; 8:10; 12:13; 15; 25:36; 40:41]
W = inv(Qmm[mom_index, mom_index]) # convert to a weight matrix
data_moments = data_moments[mom_index]

# Moment restriction on signs in mom_labels[mom_index] ----
mom_labels_keep = mom_labels[mom_index]

# Moments with positive signs
mom_labels_pos = []
mom_index_pos = Vector{Int64}()
if !isempty(mom_labels_pos)
    for mom_label in mom_labels_pos
        #print(mom_label)
        index_pos = findfirst(x -> x == mom_label, mom_labels_keep)
        push!(mom_index_pos, index_pos)
    end
else
    mom_index_pos = []
end

# Moments with negative signs
mom_labels_neg = []
mom_index_neg = Vector{Int64}()
if !isempty(mom_labels_neg)
    for mom_label in mom_labels_neg
        index_neg = findfirst(x -> x == mom_label, mom_labels_keep)
        push!(mom_index_neg, index_neg)
    end
else
    mom_index_neg = []
end
